/*
 * C
 *
 * Copyright 2016 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include "LLDISPLAY_configuration.h"
#if !(LCD_BPP == 16)

/* Includes ------------------------------------------------------------------*/

#include "LLDISPLAY_st7789h2.h"
#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

/* Defines -------------------------------------------------------------------*/

#define COPY_TASK_PRIORITY ( 12 )
#define COPY_TASK_STACK_SIZE (128) // portSTACK_TYPE is 32 bits wide, so 128 * 4 = 512 bits

#if LCD_BPP == 1

#define W 0xffff	// RGB565 white
#define B 0x0000	// RGB565 black
static const uint16_t TO_BPP16[] =
{
	B,B,B,B, W,B,B,B, B,W,B,B, W,W,B,B,
	B,B,W,B, W,B,W,B, B,W,W,B, W,W,W,B,
	B,B,B,W, W,B,B,W, B,W,B,W, W,W,B,W,
	B,B,W,W, W,B,W,W, B,W,W,W, W,W,W,W,
} ;

#elif LCD_BPP == 2

#define W 0xffff	// == 0xFFFFFF in RGB888
#define I 0xad55	// == 0xAAAAAA in RGB888
#define J 0x52aa	// == 0x555555 in RGB888
#define B 0x0000	// == 0x000000 in RGB888
static const uint16_t TO_BPP16[] =
{
	B,B, J,B, I,B, W,B, B,J, J,J, I,J, W,J,
	B,I, J,I, I,I, W,I, B,W, J,W, I,W, W,W,
} ;

#elif LCD_BPP == 4

static const uint16_t TO_BPP16[] =
{
	0x0000,	// == 0x000000 in RGB888
	0x1082,	// == 0x111111 in RGB888
	0x2104,	// == 0x222222 in RGB888
	0x3186,	// == 0x333333 in RGB888
	0x4228,	// == 0x444444 in RGB888
	0x52aa,	// == 0x555555 in RGB888
	0x632c,	// == 0x666666 in RGB888
	0x73ae,	// == 0x777777 in RGB888
	0x8c51,	// == 0x888888 in RGB888
	0x9cd3,	// == 0x999999 in RGB888
	0xad55,	// == 0xaaaaaa in RGB888
	0xbdd7,	// == 0xbbbbbb in RGB888
	0xce79,	// == 0xcccccc in RGB888
	0xdefb,	// == 0xdddddd in RGB888
	0xef7d,	// == 0xeeeeee in RGB888
	0xffff,	// == 0xffffff in RGB888
} ;
#else

#error Invalid LCD_BPP

#endif

/* Fields --------------------------------------------------------------------*/

/*
 * Semaphore used to ensure the LCD is showing the last available pixels data in
 * frame buffer.
 */
static xSemaphoreHandle wait_tearing_sem;

/*
 * Semaphore used to synchronize the pixels copy from back buffer to frame buffer
 * (performed by the copy task) and the UI framework
 */
static xSemaphoreHandle wait_copy_done_sem;

/*
 * Flush data
 */
static volatile int32_t lcd_copy_size;
static volatile int8_t* lcd_copy_addr;

/* Private functions ---------------------------------------------------------*/

static void lcd_wait_copy_done(void)
{
	xSemaphoreTake(wait_copy_done_sem, portMAX_DELAY);
}

static void lcd_wait_tearing(void)
{
	xSemaphoreTake(wait_tearing_sem, portMAX_DELAY);
}

static void lcd_notify_copy_done(void)
{
	xSemaphoreGive(wait_copy_done_sem);
}

static void lcd_notify_tearing(void)
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	portBASE_TYPE res = xSemaphoreGiveFromISR(wait_tearing_sem, &xHigherPriorityTaskWoken);
	if( xHigherPriorityTaskWoken != pdFALSE )
	{
		// Force a context switch here.
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
}

static void lcd_copy_according_bpp(uint8_t pixels)
{
	// retrieve BPP16 data
	uint16_t* data = (uint16_t*)TO_BPP16;
	
	// go to first pixel to send
	data += (pixels & 0xf) * (4 / LCD_BPP);
	
#if LCD_BPP == 1
	// send two first pixels
	*((uint16_t*)FMC_BANK1_DATA) = *data++;
	*((uint16_t*)FMC_BANK1_DATA) = *data++;
#endif
	
#if LCD_BPP <= 2
	// send first bpp2 pixel or 3rd bpp1 pixel
	*((uint16_t*)FMC_BANK1_DATA) = *data++;
#endif
	
	// send last pixel
	*((uint16_t*)FMC_BANK1_DATA) = *data;
}

static void vCopyTaskFunction(void *p_arg)
{
	while(1)
	{
		// wait tearing signal
		lcd_wait_tearing();
		int32_t size = lcd_copy_size;
		lcd_copy_size = 0;

		// copy from back buffer to LCD
		while(size >= 0)
		{
			// copy first pixel(s) of current byte
			lcd_copy_according_bpp(*lcd_copy_addr);
			
			// copy next pixel(s) of current byte
			lcd_copy_according_bpp((*lcd_copy_addr) >> 4);
			
			// go to next byte
			++lcd_copy_addr;
			
			// update counter
			size -= (8 / LCD_BPP);
		}

		// notify the copy has been done
		lcd_notify_copy_done();
	}
}

/* API -----------------------------------------------------------------------*/

void LLDISPLAY_initialization(void)
{
	lcd_copy_size = 0;
	wait_tearing_sem = xSemaphoreCreateBinary();
	wait_copy_done_sem = xSemaphoreCreateBinary();

	xTaskCreate(vCopyTaskFunction, NULL, COPY_TASK_STACK_SIZE, NULL, COPY_TASK_PRIORITY, NULL);
}

void LLDISPLAY_tearing_interrupt(void)
{
	if (lcd_copy_size > 0)
	{
		// notify tearing event
		lcd_notify_tearing();
	}	
}

void Wait_LCD(void)
{
	// wait the end of previous copy
	lcd_wait_copy_done();
}

void LLDISPLAY_copy_buffer(int32_t addr, int32_t size)
{
	// prepare copy
	lcd_copy_addr = (int8_t*)addr;
	lcd_copy_size = size;
}

#endif // (!LCD_BPP == 16)

