/*
 * C
 *
 * Copyright 2016 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include "LLDISPLAY_configuration.h"
#if LCD_BPP == 16

/* Includes ------------------------------------------------------------------*/

#include "LLDISPLAY_st7789h2.h"
#include "FreeRTOS.h"
#include "semphr.h"

/* Defines -------------------------------------------------------------------*/

/*
 * DMA used to copy pixels from back buffer to frame buffer
 */
#define DMA_INSTANCE               DMA2_Stream0
#define DMA_CHANNEL                DMA_CHANNEL_0
#define DMA_INSTANCE_IRQ           DMA2_Stream0_IRQn
#define DMA_INSTANCE_IRQHANDLER    DMA2_Stream0_IRQHandler

/* Fields --------------------------------------------------------------------*/

/*
 * DMA used to copy pixels from back buffer to frame buffer
 */
static DMA_HandleTypeDef DmaHandle;

/*
 * Semaphore used to synchronize the pixels copy from back buffer to frame buffer
 * (performed by the DMA) and the UI framework
 */
static xSemaphoreHandle wait_copy_done_sem;

/*
 * Flush data
 */
static volatile int32_t lcd_copy_size;
static volatile int16_t* lcd_copy_addr;

/* Private functions ---------------------------------------------------------*/

/*
 * Force to wait the end of previous copy
 */
static void lcd_wait_copy_done(void)
{
	xSemaphoreTake(wait_copy_done_sem, portMAX_DELAY);
}

/*
 * Notify copy has been performed
 */
static void lcd_notify_copy_done(DMA_HandleTypeDef *DmaHandle)
{
	portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
	portBASE_TYPE res = xSemaphoreGiveFromISR(wait_copy_done_sem, &xHigherPriorityTaskWoken);
	if( xHigherPriorityTaskWoken != pdFALSE )
	{
		// Force a context switch here.
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
}

/*
 * Initialize DMA without starting it
 */
static void DMA_Config(void)
{
	__HAL_RCC_DMA2_CLK_ENABLE();

	DmaHandle.Init.Channel = DMA_CHANNEL;                     /* DMA_CHANNEL_0                    */
	DmaHandle.Init.Direction = DMA_MEMORY_TO_MEMORY;          /* M2M transfer mode                */
	DmaHandle.Init.PeriphInc = DMA_PINC_ENABLE;               /* Peripheral increment mode Enable */
	DmaHandle.Init.MemInc = DMA_MINC_DISABLE;                  /* Memory increment mode Enable     */
	DmaHandle.Init.PeriphDataAlignment = DMA_PDATAALIGN_HALFWORD; /* Peripheral data alignment : Word */
	DmaHandle.Init.MemDataAlignment = DMA_MDATAALIGN_HALFWORD;    /* memory data alignment : Word     */
	DmaHandle.Init.Mode = DMA_NORMAL;                         /* Normal DMA mode                  */
	DmaHandle.Init.Priority = DMA_PRIORITY_HIGH;              /* priority level : high            */
	DmaHandle.Init.FIFOMode = DMA_FIFOMODE_DISABLE;           /* FIFO mode disabled               */
	DmaHandle.Init.FIFOThreshold = DMA_FIFO_THRESHOLD_FULL;
	DmaHandle.Init.MemBurst = DMA_MBURST_SINGLE;              /* Memory burst                     */
	DmaHandle.Init.PeriphBurst = DMA_PBURST_SINGLE;           /* Peripheral burst                 */
	DmaHandle.Instance = DMA_INSTANCE;

	HAL_DMA_Init(&DmaHandle);
	HAL_DMA_RegisterCallback(&DmaHandle, HAL_DMA_XFER_CPLT_CB_ID, lcd_notify_copy_done);
	HAL_NVIC_SetPriority(DMA_INSTANCE_IRQ, 0x0F, 0);
	HAL_NVIC_EnableIRQ(DMA_INSTANCE_IRQ);
}

/* Interrupt -----------------------------------------------------------------*/

/*
 * DMA interrupt
 */
void DMA_INSTANCE_IRQHANDLER(void)
{
	// call generic DMA interrupt
	HAL_DMA_IRQHandler(&DmaHandle);
}

/* API -----------------------------------------------------------------------*/

void LLDISPLAY_initialization(void)
{
	lcd_copy_size = 0;	// means "nothing to copy"
	wait_copy_done_sem = xSemaphoreCreateBinary();
	DMA_Config();
}

void LLDISPLAY_tearing_interrupt(void)
{
	if (lcd_copy_size > 0)
	{
		// start copy		
		HAL_DMA_Start_IT(&DmaHandle, (uint32_t)lcd_copy_addr, (uint32_t)FMC_BANK1_DATA, lcd_copy_size);
		lcd_copy_size = 0;
	}
}

/*
 * Synchronizes the display stack drawing with the display framerate.<br>
 * This method is called by the display stack before the very first drawing
 * after a flush. This allows to the display driver to wait for the right
 * moment to allow the display stack to start using the back buffer
 * in order to avoid flickering.<br>
 */
void Wait_LCD(void)
{
	// wait the end of previous copy
	lcd_wait_copy_done();
}

/*
 * Performs a flush: the content of given square in back buffer must be
 * copied into LCD buffer.
 */
void LLDISPLAY_copy_buffer(int32_t addr, int32_t size)
{
	// prepare copy; next tearing will start the copy
	lcd_copy_addr = (int16_t*)addr;
	lcd_copy_size = size;
}

#endif // LCD_BPP == 16

